
import copy
import numpy
import objloader
import time
from   util import Timer

EPSILON = 0.00000001
isequal = lambda e1, e2 : numpy.abs( e1 - e2 ) < EPSILON # floating pt equality

class AABB():
    
    def __init__( self, minpt, maxpt, faces=None ):
        self.minpt = minpt
        self.maxpt = maxpt
        self.faces = faces
        
    def __repr__( self ):
        return 'minpt = %s, maxpt = %s' % ( self.minpt, self.maxpt )
    
    def split_faces( self, plane_vector ):
        left_faces, right_faces = [], []
        return left_faces, right_faces
    
    def split(self, plane_vector ): 
        # plane_vecctor is one of i) ( x, 0.0, 0.0 ) or
        #                        ii) ( 0.0, y, 0.0 ) or
        #                       iii) ( 0.0, 0.0, z )
        split_val, dim  = plane_vector
        
        midpt_max = copy.deepcopy( self.maxpt )
        midpt_max[ dim ] = split_val
        
        midpt_min = copy.deepcopy( self.minpt )
        midpt_min[ dim ] = split_val
        
        left_faces, right_faces = None, None
        if self.faces:
            left_faces, right_faces = self.split_faces( plane_vector )
            
        LeftAABB  = AABB( copy.deepcopy( self.minpt ), midpt_max, left_faces )
        RightAABB = AABB( midpt_min, copy.deepcopy( self.maxpt ), right_faces )
        
        return [ LeftAABB, RightAABB ]
    
    def isPlanarAlongAxis( self, dim ):
        delta = numpy.absolute( self.maxpt[dim] - self.minpt[dim] )
        return delta < EPSILON
    
    def surface_area( self ):
        # xy 
        a1 = ( ( self.maxpt[0] - self.minpt[0] ) * ( self.maxpt[1] - self.minpt[1] ) ) * 2.0
        # yz
        a2 = ( ( self.maxpt[1] - self.minpt[1] ) * ( self.maxpt[2] - self.minpt[2] ) ) * 2.0
        # xz
        a3 = ( ( self.maxpt[0] - self.minpt[0] ) * ( self.maxpt[2] - self.minpt[2] ) ) * 2.0
        return a1 + a2 + a3

    def center( self ):
        return ( ( self.maxpt - self.minpt ) / 2.0 ) + self.minpt

    def intersectRay( self, ray ):

        tnear = numpy.NINF
        tfar = numpy.PINF
        
        for dim in xrange( 3 ):
            if isequal( ray.direction[dim], 0.0 ):
                if self.minpt[dim] > ray.origin[dim] or self.maxpt[dim] < ray.origin[dim]:
                    return False
            else:
                t1 = ( self.minpt[dim] - ray.origin[dim] ) / ray.direction[dim]
                t2 = ( self.maxpt[dim] - ray.origin[dim] ) / ray.direction[dim]
                if t1 > t2:
                    t1, t2 = t2, t1
                if t1 > tnear:
                    tnear = t1
                if t2 < tfar:
                    tfar = t2
                if tnear > tfar or tfar < 0: return False
        return True

    def intersectRay_old( self, ray ):
        ''' ray-box intersection using Smits algorithm, which is based on slabs 
            http://www.scratchapixel.com/lessons/3d-basic-lessons/lesson-7-intersecting-simple-shapes/ray-box-intersection/
        '''
        
        divx = 1.0 / ray.direction[0]
        divy = 1.0 / ray.direction[1]
        divz = 1.0 / ray.direction[2]

        if divx >= 0.0:
            tmin = ( self.minpt[0] - ray.origin[0] ) * divx
            tmax = ( self.maxpt[0] - ray.origin[0] ) * divx
        else:
            tmin = ( self.maxpt[0] - ray.origin[0] ) * divx
            tmax = ( self.minpt[0] - ray.origin[0] ) * divx
            
        if divy >= 0.0:
            tymin = ( self.minpt[1] - ray.origin[1] ) * divx
            tymax = ( self.maxpt[1] - ray.origin[1] ) * divx
        else:
            tymin = ( self.maxpt[1] - ray.origin[1] ) * divx
            tymax = ( self.minpt[1] - ray.origin[1] ) * divx        
        
        if tmin > tymax or tymin > tmax:
            return False, None, None
        
        if tymin > tmin:
            tmin = tymin
        if tymax < tmax:
            tmax = tymax
            
        if divz >= 0.0:
            tzmin = ( self.minpt[2] - ray.origin[2] ) * divx
            tzmax = ( self.maxpt[2] - ray.origin[2] ) * divx
        else:
            tzmin = ( self.maxpt[2] - ray.origin[2] ) * divx
            tzmax = ( self.minpt[2] - ray.origin[2] ) * divx        
        
        if tmin > tzmax or tzmin > tmax:
            return False, None, None
        
        if tzmin > tmin:
            tmin = tzmin
        if tzmax < tmax:
            tmax = tzmax
        
        return True, tmin, tmax
        
    def overlapTri( self, triangle ):
        #  use separating axis theorem to test overlap between triangle and box
        #  need to test for overlap in these directions:
        #      1) the {x,y,z}-directions (actually, since we use the AABB of the triangle
        #         we do not even need to test these)
        #      2) normal of the triangle
        #      3) crossproduct(edge from tri, {x,y,z}-direction)
        #         this gives 3x3=9 more tests
        
        boxcenter = self.center()
        boxhalfsize = ( self.maxpt - self.minpt ) / 2.0 
        #normal = numpy.zeros( 3 )
        #min = max = p0 = p1 = p2 = rad = 0.0
        
        # move everything so that the boxcenter is in (0,0,0)        
        v0 = triangle.v0 - boxcenter
        v1 = triangle.v1 - boxcenter
        v2 = triangle.v2 - boxcenter
        e0 = triangle.v1 - triangle.v0
        e1 = triangle.v2 - triangle.v1
        e2 = triangle.v0 - triangle.v2

        #
        #   do 3. test the 9 tests first (this was faster)
        
        fex = numpy.abs(e0[0])
        fey = numpy.abs(e0[1])
        fez = numpy.abs(e0[2])
        
        def axistest_x01( a, b, fa, fb ):
            p0 = a * v0[1] - b * v0[2]
            p2 = a * v2[1] - b * v2[2]
            if p0 < p2:
                minv=p0
                maxv=p2
            else:
                minv=p2
                maxv=p0
            rad = fa * boxhalfsize[1] + fb * boxhalfsize[2]
            return not ( minv > rad or maxv < -rad )

        def axistest_x2( a, b, fa, fb ):
            p0 = a * v0[1] - b * v0[2]
            p1 = a * v1[1] - b * v1[2]
            if p0 < p1:
                minv=p0
                maxv=p1
            else:
                minv=p1
                maxv=p0
            rad = fa * boxhalfsize[1] + fb * boxhalfsize[2]
            return not ( minv > rad or maxv < -rad)
        
        def axistest_y1( a, b, fa, fb ):
            p0 = -a * v0[0] + b * v0[2]
            p1 = -a * v1[0] + b * v1[2]
            if p0 < p1:
                minv=p0
                maxv=p1
            else:
                minv=p1
                maxv=p0
            rad = fa * boxhalfsize[0] + fb * boxhalfsize[2]   
            return not ( minv > rad or maxv < -rad )
    
        def axistest_y02( a, b, fa, fb ):
            p0 = -a * v0[0] + b * v0[2]
            p2 = -a * v2[0] + b * v2[2]
            if p0 < p2:
                minv=p0
                maxv=p2
            else:
                minv=p2
                maxv=p0
            rad = fa * boxhalfsize[0] + fb * boxhalfsize[2]
            return not ( minv > rad or maxv < -rad )
        
        def axistest_z0( a, b, fa, fb ): 
            p0 = a * v0[0] - b * v0[1]
            p1 = a * v1[0] - b * v1[1]
            if p0 < p1:
                minv=p0
                maxv=p1
            else:
                minv=p1
                maxv=p0
            rad = fa * boxhalfsize[0] + fb * boxhalfsize[1]
            return not ( minv > rad or maxv < -rad)
    
        def axistest_z12( a, b, fa, fb ): 
            p1 = a * v1[0] - b * v1[1]
            p2 = a * v2[0] - b * v2[1]

            if p2 < p1:
                minv=p2
                maxv=p1
            else:
                minv=p1
                maxv=p2
            rad = fa * boxhalfsize[0] + fb * boxhalfsize[1]
            return not ( minv > rad or maxv < -rad )
        
        if not axistest_x01( e0[2], e0[1], fez, fey ):
            return False
        if not axistest_y02( e0[2], e0[0], fez, fex ):
            return False
        if not axistest_z12( e0[1], e0[0], fey, fex ):
            return False
        
        fex = numpy.abs( e1[0] )
        fey = numpy.abs( e1[1] )
        fez = numpy.abs( e1[2] )
        
        if not axistest_x01( e1[2], e1[1], fez, fey ):
            return False
        if not axistest_y02( e1[2], e1[0], fez, fex ):
            return False
        if not axistest_z0( e1[1], e1[0], fey, fex ):
            return False
        
        fex = numpy.abs( e2[0] )
        fey = numpy.abs( e2[1] )
        fez = numpy.abs( e2[2] )
        
        if not axistest_x2( e2[2], e2[1], fez, fey ):
            return False
        if not axistest_y1( e2[2], e2[0], fez, fex ):
            return False
        if not axistest_z12( e2[1], e2[0], fey, fex ):
            return False
        
        #   check point 1.
        #   first test overlap in the {x,y,z}-directions
        #   find min, max of the triangle each direction, and test for overlap in
        #   that direction -- this is equivalent to testing a minimal AABB around
        #   the triangle against the AABB
        
        def findminmax( vals ):
            return vals.min(), vals.max()
        
        #   test in X-direction
        minv, maxv = findminmax( numpy.array( [ v0[0],v1[0],v2[0] ] ) )
        if minv > boxhalfsize[0] or maxv < -boxhalfsize[0]:
            return False

        #   test in Y-direction
        minv, maxv = findminmax( numpy.array( [ v0[1],v1[1],v2[1] ] ) )
        if minv > boxhalfsize[1] or maxv<-boxhalfsize[1]:
            return False
        
        #   test in Z-direction
        minv, maxv = findminmax( numpy.array( [ v0[2],v1[2],v2[2] ] ) )
        if minv > boxhalfsize[2] or maxv < -boxhalfsize[2]:
            return False
        
        #   test point 2.
        #   test if the box intersects the plane of the triangle
        #   compute plane equation of triangle: normal * x + d = 0
        normal = numpy.cross( e0, e1 )
        if not self.planeBoxOverlap( normal, v0, boxhalfsize ):
            return False
        
        return True

    def planeBoxOverlap( self, normal, tri_vert, maxbox ):
        vmin, vmax = numpy.zeros( 3 ), numpy.zeros( 3 )
        for q in xrange( 3 ):
            v = tri_vert[q]
            if normal[q] > 0.0:
                vmin[q] = -maxbox[q] - v
                vmax[q] =  maxbox[q] - v
            else:
                vmin[q] =  maxbox[q] - v
                vmax[q] = -maxbox[q] - v
        if numpy.dot( normal, vmin ) > 0.0:
            return False
        if numpy.dot( normal, vmax ) >= 0.0:
            return True
        return False

class SAHKDTree( object ):

    _IDX_LOOKUP = { 'x-axis' : 0, 'y-axis' : 1 }
    KBOXRAY = 0.026 # in msecs, 
    KT    = KBOXRAY * 2 # cost of one traversal step is testing rayhitting left/right volumes
    KI    = 0.145 # cost of intersecting one triangle,  in msecs
    LEFT  = 0
    RIGHT = 1
    EVENT_END    = 0
    EVENT_PLANAR = 1
    EVENT_START  = 2    

    def __init__( self, obj ):
        self._tree = {}
        self.triangles = obj.faces.values()
        self.bbox = self.bounding_box( self.triangles )
        print 'bounding box', self.bbox.minpt, self.bbox.maxpt                
        self.SARoot = self.bbox.surface_area()
        self.build_kdtree()
    
    def surface_area( self, triangles ):
        return sum( t.area() for t in triangles )
        
    def bounding_box( self, triangles ):
        
        maxpt = numpy.array( [ numpy.NINF, numpy.NINF, numpy.NINF ] )
        minpt = numpy.array( [ numpy.PINF, numpy.PINF, numpy.PINF ] )
        for tri in triangles:
            for x in tri.vertices():
                for i, val in enumerate( x ):
                    if val < minpt[i]:
                        minpt[i] = val
                    if val > maxpt[i]:
                        maxpt[i] = val
        
        return AABB( minpt, maxpt )
    
    def SAH(self, p, V, Nl, Nr, Np ):
        # p is the split plane location, V is the voxel, Nleft, Nright, Np are the no. of tris in left, right, p resp.
        Vl, Vr = V.split( p )
        Pl = Vl.surface_area() / self.SARoot
        Pr = Vr.surface_area() / self.SARoot
        #Nleft, Nright, Np = len( Tleft ), len( Tright ), len( Tp )
        
        if Nl == 0 or Nr == 0: 
            bias_factor = 0.8 # favour splits that cut off empty space
        else:
            bias_factor = 1.0
            
        def cost( pl, pr, nl, nr ):
            return bias_factor * ( SAHKDTree.KT + SAHKDTree.KI * ( pl * nl + pr * nr ) )
        
        cost_mid_tri_to_left_voxel  = cost( Pl, Pr, Nl + Np, Nr )
        cost_mid_tri_to_right_voxel = cost( Pl, Pr, Nl, Nr + Np )
        
        if cost_mid_tri_to_left_voxel < cost_mid_tri_to_right_voxel:
            return cost_mid_tri_to_left_voxel, SAHKDTree.LEFT
        else:
            return cost_mid_tri_to_right_voxel, SAHKDTree.RIGHT

    def clipped_AABB( self, triangle, aabb ):
        # takes and triangle, clip it and returns the AABB
        original_tri_aabb = self.bounding_box( [ triangle ] )
        
        for i in xrange( 3 ): # perform the clipping
            if original_tri_aabb.maxpt[i] > aabb.maxpt[i]:
                original_tri_aabb.maxpt[i] = aabb.maxpt[i]
            if original_tri_aabb.minpt[i] < aabb.minpt[i]:
                original_tri_aabb.minpt[i] = aabb.minpt[i]
        return original_tri_aabb

    def find_split_plane( self, triangles, aabb ):
        # find split plane for triangles contained in voxel aabb using plane sweep 
        mincost, best_split_plane, add_planar_tris_to = numpy.PINF, None, None
        event_list = []
        
        def cmp( e1, e2 ):
            equal = isequal( e1[1], e2[1] )
            if e1[1] < e2[1] or ( equal and e1[3] < e2[3] ):
                return -1
            elif equal and e1[3] == e2[3]:
                return 0
            else:
                return 1
        
        for dim in xrange( 3 ):
            for t in triangles:
                B = self.clipped_AABB( t, aabb )
                if B.isPlanarAlongAxis( dim ):
                    event_list.append( ( t, B.minpt[dim], dim, SAHKDTree.EVENT_PLANAR ) ) # t falls in the plane p
                else:
                    event_list.append( ( t, B.minpt[dim], dim, SAHKDTree.EVENT_START ) )
                    event_list.append( ( t, B.maxpt[dim], dim, SAHKDTree.EVENT_END ) )
            
            event_list.sort( cmp )
        
            Nl, Np, Nr = 0, 0, len( triangles )
            
            i = 0
            num_events = len( event_list )
            while i < num_events:
                event = event_list[i] # get the current event
                p = event[1]          # p is the coordinate value of the event in the current dimension
                pplus = pminus = pplanar = 0

                while i < num_events and isequal( p, event_list[i][1] ) and event_list[i][3] == SAHKDTree.EVENT_END:
                    pminus += 1
                    i += 1
                    
                while i < num_events and isequal( p, event_list[i][1] ) and event_list[i][3] == SAHKDTree.EVENT_PLANAR:
                    pplanar += 1
                    i += 1
                    
                while i < num_events and isequal( p, event_list[i][1] ) and event_list[i][3] == SAHKDTree.EVENT_START:
                    pplus += 1
                    i += 1
                
                Np = pplanar
                Nr -= pplanar
                Nr -= pminus
                
                split_plane = ( p, dim )
                current_cost, pside = self.SAH( split_plane, aabb, Nl, Nr, Np )
                
                if current_cost < mincost:
                    #print 'replacing mincost %s' % mincost, 
                    mincost = current_cost
                    #Print ' With new mincost %s' % mincost
                    #print 'split is at %s, dim=%s' % ( p, dim )
                    best_split_plane = ( p, dim )
                    add_planar_tris_to = pside
                    
                Nl += pplus
                Nl += pplanar
                Np = 0    
                
        return best_split_plane, add_planar_tris_to
        
    def classify( self, triangles, V, Vl, Vr, split_plane ):
        Tl, Tr, Tp = [], [], [] # holds triangle on the l, r and co-planar with the split plane
        p, dim = split_plane
        Vleft  = copy.deepcopy( Vl )
        Vright = copy.deepcopy( Vr )
        Vleft.maxpt[ dim ]  -= EPSILON
        Vright.minpt[ dim ] += EPSILON
        
        for t in triangles:
            coord_values = numpy.array( [ t.v0[dim], t.v1[dim], t.v2[dim] ] )
            test_values = numpy.array( [ p, p, p ] )
            if numpy.allclose( coord_values, test_values ) and V.overlapTri( t ):
                Tp.append( t )
            else:
                addleft, addright = False, False
                if Vleft.overlapTri( t ):
                    Tl.append( t )
                    addleft = True
                elif Vright.overlapTri( t ):
                    Tr.append( t )
                    addright = True
                    
                if addleft and addright:
                    print 'add left and add right'
                    print t
                    print Vl
                    print Vr
                    
        # sanity check
        if len( triangles ) != len( Tp ) + len( Tl ) + len( Tr ):
            raise Exception( 'Classify op lost some triangles!!!' )                        
        
        return Tl, Tr, Tp

    def intersectRay( self, ray ):

        def intersect( d, ray ):
            #print 'intersecting %s with ray %s' % ( d['box'], ray )
            if d[ 'box' ].intersectRay( ray ):
                if not d['split_plane']:
                    return d[ 'triangles' ]
                else:
                    left_node_tris = intersect( d[ 'left_node' ], ray )
                    right_node_tris = intersect( d[ 'right_node' ], ray )
                    return left_node_tris + right_node_tris
            else:
                return []

        return intersect( self.tree, ray )

    def build_kdtree( self ):

        dic = {}
        MAXLEVEL = 16
        MAXLEVEL = 4 # debug

        def pinfo( tree, level ):
            nL = tree['left_node']['tri_count']
            nR = tree['right_node']['tri_count']
            if tree[ 'tri_count' ] != nL + nR:
                print '!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! ERROR !!!!!!!!!!!!!!!!!!!!!!!!!!!'
            print 'level=', level
            print 'total triangle count=%s, sum LR=%s' % ( tree['tri_count'], nL + nR ) 
            print 'split plane', tree['split_plane']
            print 'num tris left of plane', nL  
            print 'num tris right of plane', nR



        def build( d, level, triangles, V ): # V is the bounding box of triangles 

            if len( triangles ) == 0:
                return { 'box':V, 'tri_count':0, 'split_plane':None, 'left_node':None, 'right_node':None, 'triangles':[] }
            
            if level == MAXLEVEL:
                #return triangles
                print '!!!!!!!!!!!! node has %s triangles !!!!!!!!!!!!' % len( triangles )
                return { 'box':V, 'tri_count':len( triangles ), 'split_plane':None, 
                         'left_node':None, 'right_node':None, 'triangles':triangles }
    
            # 2. find split plane and split triangle list, then recurse
            split_plane, pside = self.find_split_plane( triangles, V )            
            Vl, Vr = V.split( split_plane )
            Tl, Tr, Tp = self.classify( triangles, V, Vl, Vr, split_plane )            
            
            if pside == SAHKDTree.LEFT:
                Tl.extend( Tp )
            else:
                Tr.extend( Tp )

            d[ 'box' ] = V
            d[ 'tri_count' ] = len( triangles )
            d[ 'split_plane' ] = split_plane
            d[ 'left_node' ]   = build( {}, level+1, Tl, Vl )
            d[ 'right_node' ]  = build( {}, level+1, Tr, Vr )

            pinfo( d, level )
            return d
        self.tree = build( dic, 0, self.triangles, self.bounding_box( self.triangles ) )
        return self.tree
    
class Ray():

    def __init__( self, origin, direction ):
        self.origin = origin
        self.direction = direction
    
    def evaluate( self, t ):
        return self.origin + t * self.direction

    
    
def testboxtri():
    # test box triangle intersection
    box = AABB( numpy.array([0.0, 0.0, 0.0]), numpy.array([5.0, 5.0, 5.0]) )
    v0 = numpy.array( [1.0, 1.0, 1.0 ] )
    v1 = numpy.array( [3.0, 1.0, 1.0 ] )
    v2 = numpy.array( [3.0, 3.0, 1.0 ] )
    tri = objloader.Triangle( v0, v1, v2 )
    print box.overlapTri(tri)
    
    v0 = numpy.array( [1.0, 1.0, -1.0 ] )
    v1 = numpy.array( [3.0, 1.0, -1.0 ] )
    v2 = numpy.array( [3.0, 3.0, -1.0 ] )
    tri = objloader.Triangle( v0, v1, v2 )
    print box.overlapTri(tri)
    
    v0 = numpy.array( [1.0, 1.0, 6.0 ] )
    v1 = numpy.array( [3.0, 1.0, 6.0 ] )
    v2 = numpy.array( [3.0, 3.0, 6.0 ] )
    tri = objloader.Triangle( v0, v1, v2 )
    print box.overlapTri(tri)
    
def timeraybox():
    
              
    box = AABB( numpy.array( [0.0, 0.0, 0.0] ), numpy.array( [1.0, 1.0, 1.0] ) )
    ray = Ray(  numpy.array( [0.2, 0.0, -1.0] ), numpy.array( [0.2, 1.0, 0.0] ) )
    
    NUM = 100000
    with Timer() as t:
        for i in xrange( NUM ):
            box.intersectRay(ray)
    print 'ray/box intersection took %s msecs' % ( ( t.elapsed / NUM ) * 1000 )
    
def main():        
    objfile = open( 'teapot.obj' )
    loader = objloader.PixRayObjectReader()
    loader.read( objfile )
    kdtree = SAHKDTree( loader.currentObj )
    
    print kdtree.tree['box']

    ray = Ray(  numpy.array( [ -25.0, -25.0, -100.0] ), numpy.array( [0.0, 0.0, 1.0] ) )
    res = kdtree.intersectRay( ray ) 
    print len( res )
    return kdtree

    #print kdtree.bounding_box( kdtree.triangles.faces.values() )
    #tree = kdtree.build_kdtree()
    
#timeraybox()
#testboxtri()
#main()        
