"""
Id:          $Id:  $
Copyright:   2011 J.P. Morgan Chase & Co. Incorporated.  All rights reserved.
Type:        *** one of: [app, lib, pixie, test, util] ***
Description: Undocumented module created by Chu Ming Ng (chuming.ng@jpmorgan.com)
"""

import numpy
import numpy.linalg
from   scheduler import Computation, Scheduler, Task
import time
from   util import Timer

# Triangle ray intersection via 
# http://www.cs.virginia.edu/~gfx/Courses/2003/ImageSynthesis/papers/Acceleration/Fast%20MinimumStorage%20RayTriangle%20Intersection.pdf

class WavefrontReaderBase:
    """Wavefront OBJ/MTL reader base class.

    This is the base class for the OBJ and the MTL reader.
    """
    
    def __init__(self):
        """Constructor.
        """
        self.linenr = 0
        self.line = ""

    # read
    def read(self, f):
        """Read the content of a file.

        f is a file like object that can be used to read the content
        of the file.
        The file is read and for each keyword a handle_<keyword>() method
        is called with the data as argument (the number of arguments depends
        on the keyword). Each argument is of type str.
        A syntax error is generated if such a handler method is not available.
        The task of these handler methods is to preprocess the data, check
        for errors and invoke the final handler methods which are just called
        after the keyword.
        Before the file is read, the begin() method is called. At the end,
        the end() method is called.
        """

        self.linenr = 0
        self.begin()
        for self.line in f:
            self.linenr += 1
            line2 = self.line.strip()
            # Ignore empty lines and comments
            if line2 == "" or line2[0] in ['#', '$', '!', '@']:
                continue

            a = line2.split()
            cmd = a[0]
            args = a[1:]
            handler = getattr( self, "handle_%s"%cmd, None )
            if handler!=None:
                handler( *args )
            else:
                self.handleUnknown( cmd, args )

        self.end()

    # begin
    def begin( self ):
        """Begin reading a file.

        This method is called before the file is read.
        """
        pass

    # end
    def end( self ):
        """End of reading.

        This method is called after the file was read.
        """
        pass

    # handleUnknown
    def handleUnknown( self, cmd, arglist ):
        """Handle unknown keywords.

        cmd is the command keyword (the 1st argument in the current line)
        and arglist is the data (the remaining arguments).

        The default implementation raises a SyntaxError exception.
        """
        raise SyntaxError, "Unknown statement in line %d: %s"%( self.linenr, self.line )

class OBJReader( WavefrontReaderBase ):
    """Wavefront OBJ reader.

    This class can be used as base class to read an OBJ file.
    """
    
    def __init__( self ):
        """Constructor.
        """
        WavefrontReaderBase.__init__( self )
        self.v_count = 0
        self.vp_count = 0
        self.vt_count = 0
        self.vn_count = 0

    # read
    def read( self, f ):
        self.v_count = 0
        self.vp_count = 0
        self.vt_count = 0
        self.vn_count = 0
        WavefrontReaderBase.read(self, f)

    # Pre handler methods (they must be called "handle_<keyword>")

    def handle_mtllib( self, *files ):
        """Material library command.

        files contains the filenames.
        """
        if len( files )==0:
            raise SyntaxError, "No material library given in line %d: %s"%( self.linenr, self.line )
        self.mtllib( *files )

    def handle_usemtl( self, *name ):
        """Material name.

        (all names are concatenated)
        """
        self.usemtl( " ".join( name ) )

    def handle_g( self, *groups ):
        """Group command.

        groups are the group names that the following geometry belongs to.
        """
        if len( groups )==0:
            groups = ( "default", )
#            raise SyntaxError, "No group name given in line %d: %s"%(self.linenr, self.line)
        self.g( *groups )

    def handle_s( self, group_number ):
        """Smoothing group command.

        group_number is a string that contains the smoothing group number
        or "off".
        """
        group_number = group_number.lower()
        if group_number=="off":
            group_number = 0
        try:
            group_number = int( group_number )
        except:
            raise SyntaxError, "Invalid smoothing group number in line %d: %s"%( self.linenr, self.line )
        self.s( group_number )

    def handle_v( self, x, y, z, w=1 ):
        """Vertex definition.
        """
        self.v_count += 1
        self.v( numpy.array( [ float( x ), float( y ), float( z ), float( w ) ] ) )

    def handle_vp( self, u, v=0, w=1 ):
        """Vertex in parameter space.
        """
        self.vp_count += 1
        self.vp( numpy.array( [ float( u ), float( v ), float( w ) ] ) )

    def handle_vn( self, x, y, z ):
        """Normal.
        """
        self.vn_count += 1
        self.vn( numpy.array( [ float( x ), float( y ), float( z ) ] ) )

    def handle_vt( self, u, v=0, w=0 ):
        """Texture vertex.
        """
        self.vt_count += 1
        self.vt( numpy.array( [ float( u ), float( v ), float( w ) ] ) )

    def handle_p( self, *verts ):
        """Points."""
        
        if len( verts )==0:
            raise SyntaxError, "At least 1 vertex required in line %d: %s"%( self.linenr, self.line )

        vlist = []
        for s in verts:
            vert = int( s )
            if vert<0:
                vert = self.v_count+vert+1
            if vert==0:
                raise ValueError, "0-index in line %d: %s"%( self.linenr, self.line )
            vlist.append( vert )
            
        self.p( *vlist )     

    def handle_l( self, *verts ):
        """Line."""

        if len( verts ) < 2:
            raise SyntaxError, "At least 2 vertices required in line %d: %s"%( self.linenr, self.line )

        vlist = []
        for s in verts:
            a = s.split( "/" )
            if len( a ) == 0 or len( a ) > 2:
                raise SyntaxError, "Syntax error in line %d: %s"%( self.linenr, self.line )
            vert = int( a[0] )
            if vert < 0:
                vert = self.v_count + vert + 1
            tvert = None
            if len( a ) > 1:
                if a[1] != "":
                    tvert = int( a[1] )
                    if tvert < 0:
                        tvert = self.vt_count + tvert + 1
            if vert == 0 or tvert == 0:
                raise ValueError, "0-index in line %d: %s"%( self.linenr, self.line )
            vlist.append( ( vert, tvert ) ) 
        self.l( *vlist )

    def handle_f( self, *verts ):
        """Polygonal face.
        """
        if len( verts ) < 3:
            raise SyntaxError, "At least 3 vertices required in line %d: %s"%( self.linenr, self.line )

        vlist = []
        for s in verts:
            a = s.split( "/" )
            if len( a ) == 0 or len( a ) > 3:
                raise SyntaxError, "Syntax error in line %d: %s"%( self.linenr, self.line )
            vert = int( a[0] )
            if vert < 0:
                vert = self.v_count + vert + 1
            tvert = None
            normal = None
            if len( a ) > 1:
                if a[1]!="":
                    tvert = int( a[1] )
                    if tvert < 0:
                        tvert = self.vt_count + tvert + 1
                if len( a ) > 2 and a[2] != "":
                    normal = int( a[2] )
                    if normal < 0:
                        normal = self.vn_count + normal + 1
            if vert == 0 or tvert == 0 or normal == 0:
                raise ValueError, "0-index in line %d: %s"%( self.linenr, self.line )
            vlist.append( ( vert, tvert, normal ) )
        self.f( *vlist )

    def handle_o(self, name):
        """Object name.
        """
        self.o( name )

    def handle_bevel( self, on_off ):
        """Bevel interpolation on/off.
        """
        self.bevel( on_off )

    def handle_c_interp( self, on_off ):
        """Color interpolation on/off.
        """
        self.c_interp( on_off )

    def handle_d_interp( self, on_off ):
        """Dissolve interpolation on/off.
        """
        self.d_interp( on_off )

    def handle_lod( self, level ):
        """Level of Detail.
        """
        self.lod( int( level ) )

    def handle_shadow_obj( self, filename ):
        """Shadow object.
        """
        self.shadow_obj( filename )

    def handle_trace_obj( self, filename ):
        """Shadow object.
        """
        self.trace_obj( filename )

    # Handler methods

    def call( self, filename, *args ):
        pass

    def csh( self, cmd ):
        pass

    def mtllib( self, *files ):
        """Specification of material libraries.

        files is a sequence of file names that contain material definitions.
        """
        pass

    def usemtl( self, name ):
        """Material name.

        name is a string containing the name of the material to use for
        the following elements.
        """
        pass

    def g( self, *groups ):
        """Grouping statement.

        groups is a sequence of group names that the following geometry
        belongs to.
        """
        pass

    def s( self, group_number ):
        """Smoothing group.

        group_number is an integer containing the smoothing group number.
        Smoothing groups should be turned off if the group number is 0.
        """
        pass

    def v( self, vert ):
        """Geometric vertex.

        vert is always a vec4."""
        pass

    def vp( self, vert ):
        """A point in parameter space.

        This vertex is used for free-form curves or surfaces.
        vert is always a vec3.
        """
        pass

    def vn( self, normal ):
        """Normal vector.
        
        normal is a vec3.
        """
        pass

    def vt( self, tvert ):
        """Texture vertex.
        
        tvert is always a vec3.
        """
        pass

    def p( self, *verts ):
        """Points.

        verts is a list of vertex indices. The indices are always >0
        (negative indices are automatically converted to their
        corresponding positive indices).
        All indices are 1-based. If an index in the file was 0, an exception
        was already thrown.
        """
        print "POINT",verts

    def l( self, *verts ):
        """Line.

        verts contains 2-tuples (vert, tvert) which contains the indices
        to the vertexa and the texture vertex. tvert may be None.
        The indices are always >0 (negative indices are automatically
        converted to their corresponding positive indices).
        All indices are 1-based. If an index in the file was 0, an exception
        was already thrown.
        """
        pass

    def f( self, *verts ):
        """Polygonal face.
        
        verts contains 3-tuples (vert, tvert, normal) which contains
        the indices to the vertex, the texture vertex and the normal.
        tvert and normal may be None, otherwise the values are always >0
        (negative indices are automatically converted to their corresponding
        positive indices).
        All indices are 1-based. If an index in the file was 0, an exception
        was already thrown.
        """
        pass

    def o( self, name ):
        """Optional object name.

        name is a string containing the specified name for the elements
        following this statement.
        """
        pass

    def bevel( self, on_off ):
        pass

    def c_interp( self, on_off ):
        pass

    def d_interp( self, on_off ):
        pass

    def lod( self, level ):
        """
        level is an integer.
        """
        pass

    def shadow_obj( self, filename ):
        pass

    def trace_obj( self, filename ):
        pass

class PixRay3DObject( object ):

    def __init__( self, objname ):
        self.objname = objname
        self.vertices = []
        self.faces    = {}

    def __str__( self ):
        return '%s : %s vertices, %s faces' %( self.objname, self.vertices, self.faces )
    
    def addVertex( self, v ):
        self.vertices.append( v )

    def addFace( self, f ):
        self.faces[ f.ID ] = f

    def getTribyID( self, id ):
        self.faces.get( id, None )

class PixRayObjectReader( OBJReader ):

    def __init__( self ):
        OBJReader.__init__( self )
        self.currentObj = None
        self.objlist = []

    def g( self, groups ):
        if self.currentObj:
            self.objlist.append( self.currentObj )
        self.currentObj = PixRay3DObject( groups )
        
    def v( self, vert ):
        self.currentObj.addVertex( vert[:3] ) # drop the forth coordinate w

    def f( self, *verts ):
        v1, v2, v3 = zip(*verts)[0]
        v1, v2, v3 = [ self.currentObj.vertices[ x - 1 ] for x in [ v1, v2, v3 ] ]
        f = Triangle( v1, v2, v3 )
        self.currentObj.addFace( f )

class Triangle():

    ID = 0
    def __init__( self, v0, v1, v2 ):
        self.id = Triangle.ID
        Triangle.ID += 1
        self.v0 = v0
        self.v1 = v1
        self.v2 = v2
        self.E1 = self.v1 - self.v0
        self.E2 = self.v2 - self.v0
        self.normal = numpy.cross( self.E1, self.E2 )
        
    def __str__( self ):
        return 'Triangle : v1=%s, v2=%s, v3=%s, n=%s' % ( self.v0, self.v1, self.v2, self.normal ) 

    def vertices( self ):
        return [ self.v0, self.v1, self.v2 ]

    def area( self ):
        ''' returns area of the triangle '''
        return 0.5 * numpy.linalg.norm( self.normal ) 

    #@gpu.kernel( name='ray_tri_intersection' )
    def intersect_kernel( self ):
        return '''
        __global__ void ray_tri_intersection( float *dest, 
                                              float *rpx, float *rpy, float *rpz, 
                                              float *rdx, float *rdy, float *rdz,
                                              float *v0x, float *v0y, float *v0z,
                                              float *v1x, float *v1y, float *v1z,
                                              float *v2x, float *v2y, float *v2z, int n ) 
                                              
        {
            const int i = threadIdx.x + blockIdx.x * blockDim.x;
            if( i < n )
            {
                float3 ray_origin = make_float3( rpx[i], rpy[i], rpz[i] );
                float3 ray_direction = make_float3( rdx[i], rdy[i], rdz[i] );
                float3 v0 =  make_float3( v0x[i], v0y[i], v0z[i] );
                float3 v1 =  make_float3( v1x[i], v1y[i], v1z[i] );
                float3 v2 =  make_float3( v2x[i], v2y[i], v2z[i] );
                float3 E1 = v1 - v0;
                float3 E2 = v2 - v0;
                float3 T = ray_origin - v0;
                float3 P = cross( ray_direction, E2 );
                float PdotE1 = dot( P, E1 );
                
                if ( PdotE1 < 0.000000001f )
                    dest[i] = __int_as_float(0x7fffffff); // no hit
                    
                float3 Q = cross( T, E1 );
                float QdotE2 = dot( Q, E2 );
                float PdotT = dot( P, T );
                float QdotD = dot( Q, ray_direction );
                float3 bary_coords = make_float3( QdotE2, PdotT, QdotD ) / PdotE1; 
                if( bary_coords.x > 0.0f && ( bary_coords.y >= 0.0f && bary_coords.y <= 1.0f ) && ( bary_coords.z >= 0.0f && bary_coords.z <= 1.0f ) )
                    dest[i] = bary_coords.x;
                else
                    dest[i] = __int_as_float(0x7fffffff); // no hit
            } 
        }
        '''

    def intersect_cr( self, ray ):
        t = yield Computation( 'ray_tri_intersection', numpy.concatenate( [ ray.origin, ray.direction, self.v0, self.v1, self.v2 ] ).tolist() )
        yield ray.origin + t * ray.direction

    def intersect( self, ray ):
        # compute intersection using barycentric coordinates
        T = ray.origin - self.v0
        P = numpy.cross( ray.direction, self.E2 )
        PdotE1 = numpy.dot( P, self.E1 )
        if PdotE1 == 0.0:
            return None
        Q = numpy.cross( T, self.E1 )
        QdotE2 = numpy.dot( Q, self.E2 )
        PdotT = numpy.dot( P, T )
        QdotD = numpy.dot( Q, ray.direction )
        t, u, v = numpy.array( [ QdotE2, PdotT, QdotD ] ) / PdotE1
        #print 't, u, v = %s, %s, %s' % ( t, u, v )
        if t > 0.0 and ( u >= 0.0 and u <= 1.0 ) and ( v >= 0.0 and v <= 1.0 ):
            return ray.origin + t * ray.direction
        return None
        
    
class Ray():

    def __init__( self, origin, direction ):
        self.origin = origin
        self.direction = direction

    def __repr__( self ):
        return 'Ray origin = %s, direction = %s' % ( self.origin, self.direction ) 

    def evaluate( self, t ):
        return self.origin + t * self.direction
    
def main():
    
    v1 = numpy.array( [0.0, 0.0, 0.0 ] )
    v2 = numpy.array( [1.0, 1.0, 0.0 ] )
    v3 = numpy.array( [2.0, 0.0, 0.0 ] )
    t = Triangle( v1, v2, v3 )
    
    ray = Ray( numpy.array( [1.0, 0.5, 1.0] ), numpy.array( [0.0, 0.0, -1.0] ) )
    print t.intersect( ray )    
    
    ray2 = Ray( numpy.array( [1.0, 0.5, 1.0] ), numpy.array( [1.0, 0.0, 0.0] ) )
    print t.intersect( ray2 )
    
    objfile = open( 'teapot.obj' )
    objloader = PixRayObjectReader()
    objloader.read( objfile )
    print objloader.v_count

    import pprint
    pprint.pprint( objloader.currentObj )
    return objloader

#test_scheduler_call()
#test_explicit_call()
#main()

