
import numpy
import scipy
import StringIO
import string
import sys
import pprint
import math
import PIL.Image
import copy
import sys
import math
import time
import json
import scipy.optimize
from   functools import partial

from   optparse import OptionParser
from   util import Timer

Origin          = numpy.array( [ 0.0, 0.0, 0.0 ] )
XBasisVector    = numpy.array( [ 1.0, 0.0, 0.0 ] )
YBasisVector    = numpy.array( [ 0.0, 1.0, 0.0 ] )
ZBasisVector    = numpy.array( [ 0.0, 0.0, 1.0 ] )

NonMaterial     = numpy.array( [ 0.0, 0.0, 0.0 ] )  
ZeroVector      = numpy.array( [ 0.0, 0.0, 0.0 ] )  
OneVector       = numpy.array( [ 1.0, 1.0, 1.0 ] )  
BackgroundColor = numpy.array( [ 0.1, 0.1, 0.1 ] ) # default background color

RayCutoffDepth = 5
Epsilon = 0.1

options = None


def reflect_vector( vector, normal ):
    return vector - 2 * numpy.dot( vector, normal ) * normal

def reflected_ray( ray, normal, point ):
    new_dir = reflect_vector( ray.direction, normal )
    return Ray( point, new_dir, ray.screen_pos, (0.0, 0.0), ray.depth + 1 )

def rotation_matrix(angle, direction, point=None):
    """Return matrix to rotate about axis defined by point and direction.

    >>> R = rotation_matrix(math.pi/2, [0, 0, 1], [1, 0, 0])
    >>> numpy.allclose(numpy.dot(R, [0, 0, 0, 1]), [1, -1, 0, 1])
    True
    >>> angle = (random.random() - 0.5) * (2*math.pi)
    >>> direc = numpy.random.random(3) - 0.5
    >>> point = numpy.random.random(3) - 0.5
    >>> R0 = rotation_matrix(angle, direc, point)
    >>> R1 = rotation_matrix(angle-2*math.pi, direc, point)
    >>> is_same_transform(R0, R1)
    True
    >>> R0 = rotation_matrix(angle, direc, point)
    >>> R1 = rotation_matrix(-angle, -direc, point)
    >>> is_same_transform(R0, R1)
    True
    >>> I = numpy.identity(4, numpy.float64)
    >>> numpy.allclose(I, rotation_matrix(math.pi*2, direc))
    True
    >>> numpy.allclose(2, numpy.trace(rotation_matrix(math.pi/2,
    ...                                               direc, point)))
    True

    """
    sina = math.sin(angle)
    cosa = math.cos(angle)
    #direction = unit_vector(direction[:3])
    direction = direction / numpy.linalg.norm( direction )
    # rotation matrix around unit vector
    R = numpy.diag([cosa, cosa, cosa])
    R += numpy.outer(direction, direction) * (1.0 - cosa)
    direction *= sina
    R += numpy.array([[ 0.0,         -direction[2],  direction[1]],
                      [ direction[2], 0.0,          -direction[0]],
                      [-direction[1], direction[0],  0.0]])
    M = numpy.identity(4)
    M[:3, :3] = R
    if point is not None:
        # rotation not around origin
        point = numpy.array(point[:3], dtype=numpy.float64, copy=False)
        M[:3, 3] = point - numpy.dot(R, point)
    return M

def lightVector( point, light ):
    l = light.origin - point
    return l / numpy.linalg.norm( l )

class RedirectIO():

    def __enter__( self ):
        self.stdout = sys.stdout
        self.strio = StringIO.StringIO()
        sys.stdout = self.strio
        return self

    def __exit__(self, type, value, traceback):
        sys.stdout = self.stdout

    def getvalue( self ):
        return self.strio.getvalue()

class Light():

    def __init__( self, origin, diffuse=OneVector, specular=OneVector ):
        self.origin  = origin
        self.diffuse = diffuse
        self.specular = specular
    
    def __repr__( self ):
        return 'Light origin=%s, diffuse=%s, specular=%s' % ( self.origin, self.diffuse, self.specular )


class Ray():

    def __init__( self, origin, direction, screen_pos, coeffs, depth=0, isEyeRay=False ):
        self.origin = origin
        self.direction = direction
        self.screen_pos = screen_pos
        self.coeffs = coeffs
        self.depth = depth
        self.isEyeRay = isEyeRay

    def __repr__( self ):
        return 'Ray origin = %s, direction = %s' % ( self.origin, self.direction ) 

    def evaluate( self, t ):
        return self.origin + t * self.direction

class Hit():

    def __init__( self, point, ray, obj, t ):
        self.point = point
        self.ray = ray
        self.obj = obj
        self.t = t

    def evalLighting( self, light ):
        return self.obj.render( self.point, light )

    def __repr__( self ):
        return 'Hit => point = %s,\n\tray = %s,\n\tobj = %s\n\tt=%s' % ( self.point, self.ray, self.obj, self.t )

class SceneGraph():

    def __init__( self, objs, lights, background=None ):
        # TODO build some hierarchical data structure. use a list of objs for now
        self.objs = objs
        self.lights = lights
        self.background = background

class Material():

    def __init__( self, emission=ZeroVector, ambient =ZeroVector, diffuse =ZeroVector, 
                        specular=ZeroVector, shininess=0.0, reflective=False ):
        self.emission = emission
        self.ambient  = ambient
        self.diffuse  = diffuse
        self.specular = specular
        self.shininess = shininess
        self.reflective = reflective

class Texture():

    def __init__( self, filename ):
        self.image = PIL.Image.open( filename )
        self.pix = self.image.load()

    def getTexel( self, u, v ):

        xres, yres = self.image.size
        x, y = int( u * xres ), int( v * yres )
        # remember for PIL Image (0, 0) is the top left corner so we need to invert
        c = self.pix[ x, yres - y - 1 ]
        c = c[:3]
        return numpy.array( c, dtype=float ) / 255.0

class SurfaceRenderMixin():

    def render( self, point, lights, eye ):
        if self.material.reflective:
            return None
        return numpy.sum( [ self.render_i( point, light, eye ) for light in lights ], axis=0 )

    def render_i( self, point, light, eye ):
        # apply lighting model here
        n = self.normalAt( point )
        l = lightVector( point, light )
        
        # TODO add ambient term

        # compute diffuse term
        diffuse = max( numpy.dot( n, l ), 0 ) * light.diffuse * self.material.diffuse

        # compute the specular term
        R = reflect_vector( -l, n ) 
        V = eye - point
        V = V / numpy.linalg.norm( V )
        dp = max( numpy.dot( R, V ), 0)
        specular = ( dp ** self.material.shininess ) * light.specular * self.material.specular

        result = diffuse + specular
        return result


class Rect():

    def __init__( self, bottomleft, u, v, width, height, material=None, texture=None ):
        self.bottomleft = bottomleft
        self.u = u
        self.v = v
        self.n = numpy.cross( u, v ) # u, v should be unit vectors
        self.width = width
        self.height = height 
        self.material = material or Material()
        self.texture = texture

    def __repr__( self ):
        return 'Rect bottom left at %s, width %s height %s' % ( self.bottomleft, self.width, self.height )  
        
    def pointInRect( self, point ):
        # test that point is in rect. Use a.b = |a||b|cos(x)
        # u axis 
        w = point - self.bottomleft
        proj_on_u = numpy.dot( w, self.u ) # |self.u| = 1
        # v axis
        proj_on_v = numpy.dot( w, self.v ) # |self.u| = 1
        if ( proj_on_u > 0.0 and proj_on_u < self.width ) and ( proj_on_v > 0.0 and proj_on_v < self.height ):
            return True
        return False
            
    def intersect( self, ray ):

        if ray.isEyeRay: # don't interact with eye rays
            return None

        # find t =  - [ ( ray.origin - rect.bottomleft ) . rect.n ] / ray.direction . rect.n 
        ray_dot_n = numpy.dot( self.n, ray.direction )
        if  ray_dot_n < 0.0:
            # note : need to be careful with a divide by zero here is the 
            t = - numpy.dot( ray.origin - self.bottomleft, self.n ) / ray_dot_n
            if t > 0.0:
                hit = ray.evaluate( t )
                if self.pointInRect( hit ):
                    return Hit( hit, ray, self, t )
        return None

    def normalAt( self, point ):
        return self.n

    def render( self, point, lights, eye ):
        tex_color, lighting_color = ZeroVector, ZeroVector
        if self.texture:
            w = point - self.bottomleft
            u = numpy.dot( w, self.u ) / float( self.width ) 
            v = numpy.dot( w, self.v ) / float( self.height ) 
            tex_color = self.texture.getTexel( u, v )

        if self.material:
            pass
        if len( tex_color ) < 3 :
            print 'Error!'
        return tex_color

class Skybox():

    def __init__( self, filenames, bottomleft, topright ):
        self.bottomleft = bottomleft
        self.topright = topright
        self.filenames = filenames # files should be in the order of top, bottom, front, back, left, right
        self.faces = []

        #Rect( bottomleft, u, v, width, height )

        bx, by, bz = bottomleft
        tx, ty, tz = topright

        # do back face 
        rect_bottom_left = numpy.array( [tx, by, tz] )
        rect_u, rect_v = numpy.array( [-1.0, 0.0, 0.0] ), numpy.array( [0.0, 1.0, 0.0] )
        rect_width, rect_height = tx - bx, ty - by
        self.faces.append( Rect( rect_bottom_left, rect_u, rect_v, rect_width, rect_height,
                                 texture = Texture( filenames[0] ) ) )
        # do front face
        rect_bottom_left = numpy.array( [bx, by, bz] )
        rect_u, rect_v = numpy.array( [1.0, 0.0, 0.0] ), numpy.array( [0.0, 1.0, 0.0] )
        #width and height remains the same
        self.faces.append( Rect( rect_bottom_left, rect_u, rect_v, rect_width, rect_height,
                                 texture = Texture( filenames[1] ) ) )

        # do left face
        rect_bottom_left = numpy.array( [bx, by, tz] )
        rect_u, rect_v = numpy.array( [0.0, 0.0, -1.0] ), numpy.array( [0.0, 1.0, 0.0] )
        rect_width, rect_height = tz - bz, ty - by
        self.faces.append( Rect( rect_bottom_left, rect_u, rect_v, rect_width, rect_height,
                                 texture = Texture( filenames[2] ) ) )     

        # do right face
        rect_bottom_left = numpy.array( [tx, by, bz] )
        rect_u, rect_v = numpy.array( [0.0, 0.0, 1.0] ), numpy.array( [0.0, 1.0, 0.0] )
        #width and height remains the same
        self.faces.append( Rect( rect_bottom_left, rect_u, rect_v, rect_width, rect_height,
                                 texture = Texture( filenames[3] ) ) )                                 

        # do top face
        rect_bottom_left = numpy.array( [tx, ty, tz] )
        rect_u, rect_v = numpy.array( [-1.0, 0.0, 0.0] ), numpy.array( [0.0, 0.0, -1.0] )
        rect_width, rect_height = tx - bx, tz - bz
        self.faces.append( Rect( rect_bottom_left, rect_u, rect_v, rect_width, rect_height,
                                 texture = Texture( filenames[4] ) ) )             
                   
        # do bottom face
        rect_bottom_left = numpy.array( [bx, by, tz] )
        rect_u, rect_v = numpy.array( [1.0, 0.0, 0.0] ), numpy.array( [0.0, 0.0, -1.0] )
        #width and height remains the same
        self.faces.append( Rect( rect_bottom_left, rect_u, rect_v, rect_width, rect_height,
                                 texture = Texture( filenames[5] ) ) )              
        print 'Skybox faces in back, front, left, right, top, bottom order'
        for rect in self.faces:
            print rect.bottomleft, rect.n
            
class Torus( SurfaceRenderMixin ):

    def __init__( self, R, r, center = numpy.array( [0.0, 0.0, 0.0] ), material=None, transform=numpy.identity(3) ):
        self.R = R
        self.r = r
        self.position = center
        self.material = material
        self.transform = transform
        # compute inverse once
        self.inv_matrix = numpy.linalg.inv( self.transform )
        self.inv_transpose_matrix = numpy.linalg.inv( self.transform ).transpose()

    def __repr__( self ):
        return 'Torus at center %s, outer radius %s, inner radius %s' % ( self.position, self.R, self.r )

    def intersect( self, ray ):

        # transform the ray by the translation implied by the center of the torus
        new_origin = copy.deepcopy( ray.origin )
        new_origin -= self.position
        new_origin = numpy.dot( self.inv_matrix, new_origin )

        # transform the ray direction by the inverse of the rotation transform
        new_dir = copy.deepcopy( ray.direction )
        new_dir = numpy.dot( self.inv_matrix, new_dir )

        ddotd = numpy.dot( new_dir, new_dir )
        odotd = numpy.dot( new_origin, new_dir )
        odoto = numpy.dot( new_origin, new_origin )
        rayz = new_dir[2]
        oz = new_origin[2]
        rsqr = self.r * self.r
        Rsqr = self.R * self.R
        a4 = ddotd * ddotd
        a3 = 4 * ddotd * odotd
        a2 = 4 * odotd * odotd + 2 * ddotd * ( odoto - rsqr - Rsqr ) + 4 * Rsqr * rayz * rayz
        a1 = 4 * odotd * ( odoto - rsqr - Rsqr ) + 8 * Rsqr * oz * rayz
        c = ( odoto - rsqr - Rsqr )
        fourRsqr = 4 * Rsqr
        a0 = c**2 + fourRsqr * ( oz * oz - rsqr )

        coeffs = [a4, a3, a2, a1, a0]
        poly = scipy.poly1d( coeffs )
        roots = numpy.roots( poly )

        # find all the real roots
        indices = numpy.logical_and( roots > 0.0, numpy.isreal( roots ) )
        roots = roots[ indices ]

        if roots.size > 0:
            k = numpy.min( roots )
            if k > 0.0 :
                hit = ray.evaluate( k )
                return Hit( hit, ray, self, k )

        return None # no intersection

    def normalAt( self, point ):
        new_point = copy.deepcopy( point )
        new_point -= self.position
        new_point = numpy.dot( self.inv_matrix, new_point )
        x, y, z = new_point
        rsqr = self.r**2
        Rsqr = self.R**2
        commonterm = 4 * ( x**2 + y**2 + z**2 - rsqr - Rsqr )
        n = numpy.array( [x * commonterm, y * commonterm, z * commonterm + 8 * z * Rsqr] )
        
        return numpy.dot( self.inv_transpose_matrix, n / numpy.linalg.norm( n ) )

class Sphere( SurfaceRenderMixin ):

    def __init__( self, center, radius, material ):
        self.center = center
        self.radius = radius
        self.material = material

    def __repr__( self ):
        return 'Sphere at %s, radius = %s' % ( self.center, self.radius ) 
    
    def intersect( self, ray, firsthit=True ):
        # ray sphere intersection
        # ray equation is l = a + k b
        # sphere is (x - o)^2 = r^2
        # => ( a + kb - o)^2 = r^2
        A = numpy.dot( ray.direction, ray.direction )
        t = ( ray.origin - self.center )
        B = 2 * numpy.dot( ray.direction, t )
        C = numpy.dot( t, t ) - ( self.radius )**2
        
        # k = -b +- sqrt( b^2 - 4ac ) / 2a
        sqrt_param = B**2 - ( 4 * A * C )
        if sqrt_param > 0.0:
            sqrt_term = math.sqrt( sqrt_param )
            denom = 2 * A
            kvalues = [ ( -B + sqrt_term ) / denom, ( -B - sqrt_term ) / denom ]
            if firsthit:
                k = min( kvalues )
                if k > 0.0 :
                    hit = ray.evaluate( k )
                    return Hit( hit, ray, self, k )
            else:
                return [ Hit( ray.evaluate( k ), ray, self, k ) for k in kvalues if k > 0.0  ]
        else:
            return None # no intersection

    def normalAt( self, point ):
        n =  point - self.center
        return n / numpy.linalg.norm( n )

class Metaballs( SurfaceRenderMixin ):

    class Ball():
        def __init__( self, position, radius ):
            self.position = position
            self.radius = radius

    def __init__( self, balls, level, material ):
        self.balls = balls
        self.level = level
        self.material = material

    def __repr__( self ):
        l = []
        for x in self.balls:
            l.append( '\tball position %s, radius %s' % ( x.position, x.radius ) )
        return '\n'.join( l )

    def potential( self, ball, point ):
        # computes the contribution to potential for ball at point
        # p( x, y, z ) = ( 1 - ( |v| / ball.radius ) ^ 2 ) ^ 3
        v = point - ball.position

        vmod = numpy.sqrt( numpy.dot( v, v ) )
        if vmod > ball.radius:
            p = 0.0
        else:
            p = ( 1 - ( vmod / ball.radius )**2 )**3

        return p

    def fn( self, point ):
        # evaluate the function at point x
        s = sum( [ self.potential( b, point ) for b in self. balls ] )
        return s - self.level

    def fn_derivative( self, point ):
        
        grad = numpy.array( [0.0, 0.0, 0.0] )
        for b in self.balls:
            v = point - b.position
            rsqr = b.radius**2
            vmod = numpy.dot( v, v )

            if numpy.sqrt( vmod ) > b.radius:
                continue

            const_term =  -6.0 *  ( ( vmod / rsqr ) - 1 )**2 / rsqr
            # remember that the gradient points to the direction of greatest increase in potential
            # hence we need to take a negative of that to get the normal that is pointing outwards.
            grad += -( v * const_term )
        return grad

    def intersect( self, ray ):
        
        # create the spheres for the balls and bound spheres for the balls
        spheres = []
        for x in self.balls:
            s = Sphere( x.position, x.radius, None )
            spheres.append( s )
            s = Sphere( x.position, x.radius * 2.0, None )
            spheres.append( s )

        hits = []
        for x in spheres:
            intersections = x.intersect( ray, firsthit=False ) 
            if intersections:
                hits.extend( intersections )

        hits = filter( None, hits )
        if hits:
            hits.sort( key=lambda x : x.t )
            hit = hits[0]
            k = hit.t
            lastk = hits[-1].t
            fval = self.fn( hit.ray.evaluate( k ) )
            found = True
            while fval < 0.0:
                # march along the ray until fval turns positive
                k += 1.0
                pt = hit.ray.evaluate( k ) 
                fval = self.fn( pt )
                if k > lastk:
                    # we marched past the exit point so terminate
                    found=False
                    #print 'break out of march'
                    break

            if found:
                # use secant method to find to exact root

                def f( t ):
                    return self.fn( hit.ray.evaluate( t ) )
                    
                exact_k = scipy.optimize.newton( f, k )
                exact_pt = hit.ray.evaluate( exact_k )
                return Hit( exact_pt, hit.ray, self, exact_k ) 
        return None

    def normalAt( self, point ):
        n = self.fn_derivative( point )
        n = n / numpy.linalg.norm( n )
        return n

class Camera():

    def __init__( self,  eye, u, v, w, image_distance, screen ):
        self.eye = eye
        self.u = u # u, v, w are the orthonormal basis
        self.v = v
        self.w = w 
        self.image_distance = image_distance
        self.screen = screen

        # derive the value for c, the lower left corner of the screen. 
        # c = eye - image_distance * w - 0.5 * screen.width * u - 0.5 * screen.height * v
        self.c = eye - ( image_distance * w ) - ( 0.5 * screen.width * u ) - ( 0.5 * screen.height * v )

        # derive the value for b, the vector from c along the direction of v
        self.b = ( screen.height * v )

        # derive the value for a, the vector from c along the direction of u
        self.a = ( screen.width * u )

    def __repr__( self ):

        with RedirectIO() as r:
            print 'Camera configuration,'
            print '\teye is at %s' % self.eye
            print '\timage distance is %s' % self.image_distance        
            print '\tscreen width, height = %s, %s' % ( self.screen.width, self.screen.height )
            print '\tv is %s' % self.v
            print '\tu is %s' % self.u
            print '\tw is %s' % self.w
            print '\tc is at %s' % self.c
            print '\tb is %s' % self.b
            print '\ta is %s' % self.a

        return r.getvalue()

    def getDirectionVector( self, alpha, beta ):
        v = ( self.c + alpha * self.a + beta * self.b ) - self.eye
        return v / numpy.linalg.norm( v )

    def generateRays( self ):

        # shoot ray thru each pixel
        alphas = numpy.array( xrange(0, self.screen.xres, 1) )
        betas  = numpy.array( xrange(0, self.screen.yres, 1) )
        #rays = []
        for x in alphas:
            for y in betas:
                xx = float( x ) / self.screen.xres
                yy = float( y ) / self.screen.yres
                direction = self.getDirectionVector( xx, yy )
                yield Ray( self.eye, direction, ( x, y ), ( xx, yy ), 0, isEyeRay=True )

class Renderer():

    def __init__( self, camera, scene ):
        self.camera = camera
        self.scene = scene

    def supersample( self, ray, obj ):

        sw = 1.0 / ( 2.0 * self.camera.screen.xres )        
        sh = 1.0 / ( 2.0 * self.camera.screen.yres )
        pixel_spacing = numpy.array( [sw, sh] )
        coeffs = numpy.array( [ [ -1.0, 0.0 ], [ 1.0, 0.0 ], [ 0.0, 1.0 ], [ 0.0, -1.0 ] ] )
        final_coeffs = ( coeffs * pixel_spacing ) + numpy.array( ray.coeffs )
        dir_vectors = [ self.camera.getDirectionVector( i, j ) for i, j in final_coeffs ]
        rays = [ Ray( self.camera.eye, v, ray.screen_pos, (0.0, 0.0) ) for v in dir_vectors ]
        a = [ self.trace( r, [ obj ], supersample=False ) for r in rays ]
        return numpy.mean( a, axis=0 ) # take the average of the colors

    def trace( self, ray, objs, supersample=False ):
        hits = filter( None, [ obj.intersect( ray ) for obj in objs ] )
        color = BackgroundColor
        if hits:
            nearest_hit = hits[0]
            for h in hits[1:]:
                if h.t < nearest_hit.t:
                    nearest_hit = h

            colors = []
            if nearest_hit.obj.material.reflective and ray.depth < RayCutoffDepth:
                # compute relected ray and call trace again
                reflected = reflected_ray( ray, nearest_hit.obj.normalAt( nearest_hit.point ), nearest_hit.point ) 
                reflected.origin += Epsilon * reflected.direction 
                reflected_colors = self.trace( reflected, objs )
                colors.append( reflected_colors )
            else:    
                light_contribs = nearest_hit.obj.render( nearest_hit.point, self.scene.lights, self.camera.eye ) 
                if light_contribs is not None:
                    colors.append( light_contribs )

            if supersample:
                colors.append( self.supersample( ray, nearest_hit.obj ) )

            if colors:
                color = numpy.mean( colors, axis=0 )
            
        return color

    def render( self ):
        rays = self.camera.generateRays()

        if self.scene.background is not None:
            background_image = PIL.Image.open( self.scene.background )
            image = background_image.resize( ( self.camera.screen.xres, self.camera.screen.yres ) )
        else:
            image = PIL.Image.new( 'RGB', ( self.camera.screen.xres, self.camera.screen.yres ) )
        total_rays = self.camera.screen.xres * self.camera.screen.yres 
        print 'generated %s rays' % total_rays
        count = 0
        interval = int( 0.01 * total_rays )

        for r in rays:
            count += 1
            color = self.trace( r, self.scene.objs, supersample=options.supersample )
            x, y = r.screen_pos   
            #print x, y
            if count % interval == 0:
                print '.',
                sys.stdout.flush()
            if id( color ) != id( BackgroundColor ):
                pos = ( x, self.camera.screen.yres - y -1 )
                image.putpixel( pos, tuple( ( 255 * color ).astype( int ) ) )
        print        
        
        image.save( './rc_out.png' )

class Screen():

    # TODO : to wrap pygame's screen buffer objects
    def __init__( self, width, height, xres, yres ):
        self.width = width
        self.height = height
        self.xres = xres
        self.yres = yres
        if options.fast:
            self.xres = 320
            self.yres = 200
            
        print self.width, self.height, self.xres, self.yres
         
class SceneFile():

    def __init__( self, filename ):
        if not filename:
            raise Exception()
        self.filename = filename
        self.load(filename)

    def load( self, filename ):
        f = open( filename )
        s = f.read()
        f.close()

        obj = json.loads( s )

        print 'Scene file...'
        background = None
        if 'background' in obj.keys():
            background = obj['background']
            print 'background=%s' % background

        screen_data = obj['screen']
        screen = Screen( screen_data['width'],
                         screen_data['height'],
                         screen_data['xres'],
                         screen_data['yres'],
                         )
        print 'image resolution=', screen.xres,'x', screen.yres

        camera_data = obj['camera']
        args = [camera_data['position'], 
                camera_data['u'], 
                camera_data['v'], 
                camera_data['w']]
        args = [ numpy.array( x ) for x in args ]
        args.append( camera_data['image_distance'] )
        args.append( screen ) 
        self._camera = Camera( *args )

        print self._camera
        objects = []
        for x in obj['objects']:
            result = self.processObject( x )
            if isinstance( result, list ):
                objects.extend( result )
            else:
                objects.append( result )

        pprint.pprint( objects )

        lights = []
        if 'lights' in obj.keys():
            lights = [ self.processLights( x ) for x in obj['lights'] ]
            
        print 'Lights...'
        pprint.pprint( lights )
        self._scene_graph = SceneGraph( objects, lights, background )

    def processLights( self, data ):
        print 'processLights'
        origin = data['origin']
        args = [ origin ]
        if 'diffuse' in data.keys():
            diffuse = numpy.array( data['diffuse'] )
            args.append( diffuse )
        if 'specular' in data.keys():
            specular = numpy.array( data['specular'] )
            args.append( specular )
        return Light( *args )

    def processObject( self, objdata):
        return getattr( self, 'process'+objdata['type'] )( objdata )

    def processSphere( self, data ):
        position = numpy.array( data['position'] )
        radius = data['radius']
        return Sphere( position, radius, self.processMaterial( data['material'] ) )

    def processMaterial( self, data ):
        diffuse = numpy.array( data['diffuse'] )
        specular = numpy.array( data['specular'] )
        shininess = data['shininess']
        reflective = False
        if 'reflective' in data.keys():
            reflective = data['reflective']
        return Material( diffuse = diffuse, specular = specular, 
                         shininess = shininess, reflective=reflective )

    def processTorus( self, data ):
        position = numpy.array( data['position'] )
        major_radius = data[ 'major_radius' ]
        minor_radius = data[ 'minor_radius' ]
        material = self.processMaterial( data[ 'material' ] )
        transforms = [ self.processTransform( x ) for x in data[ 'transforms' ]  ]
        matrix = self.concatenateTransforms( transforms )
        print 'matrix = \n\t', matrix
        return Torus( major_radius, minor_radius, position, material, matrix )

    def concatenateTransforms( self, transforms ):
        matrix = numpy.identity( 3 )
        for x in transforms:
            matrix = numpy.dot( x, matrix )        
        return matrix

    def processTransform( self, data ):
        angle = eval( data['rotation'] )
        axis = numpy.array( data['axis'] )
        print angle, axis
        return rotation_matrix(angle, axis)[:3, :3]

    def processSkybox( self, data ):
        lower_left = numpy.array( data['lower_left'] )
        top_right = numpy.array( data['top_right'] )
        filenames = data['filenames']
        return Skybox( filenames, lower_left, top_right ).faces

    def processMetaballs( self, data ):
        l = []
        level = data['level']

        matrix = numpy.identity( 3 )
        if 'transforms' in data.keys():
            transforms = [ self.processTransform( x ) for x in data[ 'transforms' ]  ]
            matrix = self.concatenateTransforms( transforms )
            print 'transforms = %s' % transforms
            print 'matrix = %s' % matrix

        # TODO : put this into config file, hardcode for now.
        translate = numpy.array( [ 0.0, 0.0, 500 ] ) 


        for x in data['balls']:
            p = numpy.dot( matrix, numpy.array( x['position'] ) + translate ) - translate
            r = x['radius']
            l.append( Metaballs.Ball( p, r ) )
        material = self.processMaterial( data[ 'material' ] )

        return Metaballs( l, level, material )

    def getSceneGraph( self ):
        return self._scene_graph

    def getCamera( self ):
        return self._camera
   
    scene_graph = property( getSceneGraph )
    camera = property( getCamera )

def main():

    print 'Raycasting demo in python...'    
    parser = OptionParser()
    parser.add_option("-f", "--fast",
                      action="store_true", dest="fast", default=False,
                      help="fast tracing using 320x200 resolution")

    parser.add_option("-j", "--supersample",
                      action="store_true", dest="supersample", default=False,
                      help="turn on super sampling")

    parser.add_option("-s", "--scene", dest="scene_file", default='',
                      help="turn on super sampling")
    global options
    options, args = parser.parse_args()
    if not options.scene_file:
	parser.print_help()
	sys.exit()

    print 'rendering with fast=%s, supersample=%s' % ( options.fast, options.supersample )
    scene_file = SceneFile( options.scene_file )
    renderer = Renderer( scene_file.camera, scene_file.scene_graph )
    with Timer() as t:
        renderer.render()
        
    print 'Render complete, time taken = %s secs' % t.elapsed

if __name__ == '__main__':
    main()

