
import numpy
import gpu
import pycuda.driver as drv
from   scheduler import Computation, Scheduler, Task
import time
from   util import Timer

class Triangle():

    def __init__( self, v0, v1, v2 ):
        self.v0 = v0
        self.v1 = v1
        self.v2 = v2
        self.E1 = self.v1 - self.v0
        self.E2 = self.v2 - self.v0
        self.normal = numpy.cross( v1 - v0, v2 - v0 )
    
    def vertices( self ):
        return [ self.v0, self.v1, self.v2 ]
    
    def __str__( self ):
        return 'Triangle : v1=%s, v2=%s, v3=%s, n=%s' % ( self.v0, self.v1, self.v2, self.normal ) 

    @gpu.kernel( name='ray_tri_intersection' )
    def intersect_kernel( self ):
        return '''
        __global__ void ray_tri_intersection( float *dest, 
                                              float *rpx, float *rpy, float *rpz, 
                                              float *rdx, float *rdy, float *rdz,
                                              float *v0x, float *v0y, float *v0z,
                                              float *v1x, float *v1y, float *v1z,
                                              float *v2x, float *v2y, float *v2z, int n ) 
                                              
        {
            const int i = threadIdx.x + blockIdx.x * blockDim.x;
            if( i < n )
            {
                float3 ray_origin = make_float3( rpx[i], rpy[i], rpz[i] );
                float3 ray_direction = make_float3( rdx[i], rdy[i], rdz[i] );
                float3 v0 =  make_float3( v0x[i], v0y[i], v0z[i] );
                float3 v1 =  make_float3( v1x[i], v1y[i], v1z[i] );
                float3 v2 =  make_float3( v2x[i], v2y[i], v2z[i] );
                float3 E1 = v1 - v0;
                float3 E2 = v2 - v0;
                float3 T = ray_origin - v0;
                float3 P = cross( ray_direction, E2 );
                float PdotE1 = dot( P, E1 );
                
                if ( PdotE1 < 0.000000001f )
                    dest[i] = __int_as_float(0x7fffffff); // no hit
                    
                float3 Q = cross( T, E1 );
                float QdotE2 = dot( Q, E2 );
                float PdotT = dot( P, T );
                float QdotD = dot( Q, ray_direction );
                float3 bary_coords = make_float3( QdotE2, PdotT, QdotD ) / PdotE1; 
                if( bary_coords.x > 0.0f && ( bary_coords.y >= 0.0f && bary_coords.y <= 1.0f ) && ( bary_coords.z >= 0.0f && bary_coords.z <= 1.0f ) )
                    dest[i] = bary_coords.x;
                else
                    dest[i] = __int_as_float(0x7fffffff); // no hit
            } 
        }
        '''

    def intersect_gpu( self, ray ):
        t = yield Computation( 'ray_tri_intersection', numpy.concatenate( [ ray.origin, ray.direction, self.v0, self.v1, self.v2 ] ).tolist() )
        yield ray.origin + t * ray.direction

    def intersect( self, ray ):
        # compute intersection using barycentric coordinates
        T = ray.origin - self.v0
        P = numpy.cross( ray.direction, self.E2 )
        PdotE1 = numpy.dot( P, self.E1 )
        if PdotE1 == 0.0:
            return None
        Q = numpy.cross( T, self.E1 )
        QdotE2 = numpy.dot( Q, self.E2 )
        PdotT = numpy.dot( P, T )
        QdotD = numpy.dot( Q, ray.direction )
        t, u, v = numpy.array( [ QdotE2, PdotT, QdotD ] ) / PdotE1
        #print 't, u, v = %s, %s, %s' % ( t, u, v )
        if t > 0.0 and ( u >= 0.0 and u <= 1.0 ) and ( v >= 0.0 and v <= 1.0 ):
            return ray.origin + t * ray.direction
        return None
        
    
class Ray():

    def __init__( self, origin, direction ):
        self.origin = origin
        self.direction = direction

    def __repr__( self ):
        return 'Ray origin = %s, direction = %s' % ( self.origin, self.direction ) 

    def evaluate( self, t ):
        return self.origin + t * self.direction
    
def test_scheduler_call():
       
    v1 = numpy.array( [0.0, 0.0, 0.0 ] ).astype( numpy.float32 )
    v2 = numpy.array( [1.0, 1.0, 0.0 ] ).astype( numpy.float32 )
    v3 = numpy.array( [2.0, 0.0, 0.0 ] ).astype( numpy.float32 )
    
    t = Triangle( v1, v2, v3 )
    ray = Ray( numpy.array( [1.0, 0.5, 1.0]  ).astype( numpy.float32 ), 
               numpy.array( [0.0, 0.0, -1.0] ).astype( numpy.float32 ) )
    
    mod = gpu.init()
    for n in [2000]:
#    for n in [ 100, 1000, 2000, 5000, 10000, 20000, 50000, 100000, 200000, 500000, 1000000 ]:
        print 'Computing %s ray-triangle intersections...' % n
        with Timer() as timer:
            res1 = [ t.intersect( ray ) for i in xrange( n ) ]
        
        print 'CPU code took : %s secs' % timer.elapsed        

        scheduler = Scheduler( GPUKernels=mod )
        scheduler.addTasks( [ Task( t.intersect_gpu( ray ) ) for _ in xrange( n ) ] )
        with Timer() as timer:
            res2 = scheduler.run()
        print 'GPU coroutine execution model took : %s secs' % timer.elapsed        
    
        # sanity checks
        #print res1
        #print res2
        print all( map( lambda x, y : all( x == y ), res1, res2 ) )

def test_raytri_cost():
    
    v1 = numpy.array( [0.0, 0.0, 0.0 ] ).astype( numpy.float32 )
    v2 = numpy.array( [1.0, 1.0, 0.0 ] ).astype( numpy.float32 )
    v3 = numpy.array( [2.0, 0.0, 0.0 ] ).astype( numpy.float32 )
    
    t = Triangle( v1, v2, v3 )
    ray = Ray( numpy.array( [1.0, 0.5, 1.0]  ).astype( numpy.float32 ), 
               numpy.array( [0.0, 0.0, -1.0] ).astype( numpy.float32 ) )
    n=2000
    with Timer() as timer:
        for i in xrange( n ):
            t.intersect( ray ) 
        
    print 'CPU code took : %s secs' % ( timer.elapsed / n ) 
            
#test_scheduler_call()

test_raytri_cost()
