

import operator
import types
import numpy

class Computation():
    __slots__ = [ 'func_name', 'args', 'result_index' ]
    def __init__( self, func_name, args ):
        self.func_name = func_name
        self.args = args
        self.result_index = -1

def compute( c ):
    if c.func_name == 'add':
        return sum( c.args )
    elif c.func_name == 'mul':
        return reduce( operator.mul, c.args )

class Scheduler():
    
    def __init__( self, useGPU=True, GPUKernels=None ):
        self.lastReturnValues = []
        self.tasksobjs = []
        self.tasks = []
        self.useGPU = useGPU
        self.GPUKernels = GPUKernels
        
    def addTasks( self, task ):
        if isinstance( task, ( tuple, list ) ):
            self.tasksobjs.extend( task )
            self.tasks.extend( t.run() for t in task )
        else:
            self.tasksobjs.append( task )
            self.tasks.append( task.run() )
        
    def schedule_local( self, computations ):
        return [ compute( x ) for x in computations ]
        
    def schedule_gpu( self, computations ):
        # group computations into similar types
        computation_groups = {}
        for i, c in enumerate( computations ):
            c.result_index = i
            computation_groups.setdefault( c.func_name, [] ).append( c )
        
        # setup the result array
        results = [ None ] * len( computations )
        # send each group to the GPU
        for k, v in computation_groups.iteritems():
            
            args = [ numpy.array( x ).astype( numpy.float32 ) for x in zip( *[ x.args for x in v ] ) ]
            
            dest = numpy.zeros_like( args[0] )
            blocksize = 256 
            vector_size = len( dest )
            # TODO: fix up
            #args = [ drv.Out( dest ) ] + [ drv.In( x ) for x in args ] + [ drv.In( numpy.int32( vector_size ) ) ]
             
            gpu_function = self.GPUKernels.get_function( k )
            gpu_function( *args, block=( blocksize, 1, 1 ), grid=( ( vector_size + blocksize - 1 ) / blocksize, 1 ) )
            
            for r, c in zip( dest, v ):
                results[ c.result_index ] = r
                
        return results
    
    def schedule( self, computations ):
        if self.useGPU:
            return self.schedule_gpu( computations )
        else:
            return self.schedule_local( computations )
    
    def run( self ):
        # initialize None values used for first invocation of generators in self.tasks
        self.lastReturnValues = [ None ] * len( self.tasks )
        while self.tasks:
            tocompute = []
            nextpass_tasks = []
            # collect all the computations 
            for retvalue, target in zip( self.lastReturnValues, self.tasks ):
                try:
                    tocompute.append( target.send( retvalue ) )
                    nextpass_tasks.append( target )
                except StopIteration:
                    pass
            self.tasks = nextpass_tasks
            # (gpu)compute them and put then back in lastReturnValues so they will be passed back to the coroutines
            self.lastReturnValues = self.schedule( tocompute )
        return [ x.result for x in self.tasksobjs ]

class Task():
    
    def __init__( self, target ):
        self.target = target
        self.lastReturnValue = None
        self.stack = []
        self.result = None
        
    def run( self ):
        # this is the main trampolined executor. See http://en.wikipedia.org/wiki/Trampoline_%28computing%29 section on High Level Programming
        while True:
            try:
                result = self.target.send( self.lastReturnValue )
                if isinstance( result, Computation ):
                    self.lastReturnValue = yield result
                elif isinstance( result, types.GeneratorType ):
                    self.stack.append( self.target )
                    self.lastReturnValue = None
                    self.target = result
                else:
                    self.lastReturnValue = result                
            except StopIteration:
                if not self.stack:
                    #print 'Finished computing %s, result=%s' % ( self.target, self.lastReturnValue ) 
                    self.result = self.lastReturnValue
                    break
                self.target = self.stack.pop()


